@@ -20,10 +20,13 @@
 
 #include "diskscan.h"
 #include "verbose.h"
+#include "disk.h"
 #include "arch.h"
 #include "median.h"
 #include "compiler.h"
 #include "data.h"
+#include "libscsicmd/include/smartdb.h"
+#include "libscsicmd/include/ata_smart.h"
 
 #include <sched.h>
 #include <memory.h>
@@ -38,6 +41,8 @@
 #include <errno.h>
 #include <assert.h>
 
+#define TEMP_THRESHOLD 65
+
 struct scan_state {
 	uint32_t latency_bucket;
 	uint64_t latency_stride;
@@ -50,10 +55,37 @@ struct scan_state {
 	unsigned num_unknown_errors;
 };
 
+typedef int spinner_t;
+
+static char spinner_form[] = {'|', '/', '-', '\\', '|', '/', '-', '\\'};
+
+static void spinner_init(spinner_t *spinner)
+{
+	printf("%c\r", spinner_form[0]);
+	*spinner = 1;
+	fflush(stdout);
+}
+
+static void spinner_update(spinner_t *spinner)
+{
+	printf("\r%c\r", spinner_form[*spinner]);
+	if (++(*spinner) == ARRAY_SIZE(spinner_form))
+		*spinner = 0;
+	fflush(stdout);
+}
+
+static void spinner_done(void)
+{
+	printf("\r                           \r");
+	fflush(stdout);
+}
+
 const char *conclusion_to_str(enum conclusion conclusion)
 {
 	switch (conclusion) {
-		case CONCLUSION_FAILED: return "failed";
+		case CONCLUSION_FAILED_IO_ERRORS: return "failed due to IO errors";
+		case CONCLUSION_FAILED_MAX_LATENCY: return "failed due to a high max latency";
+		case CONCLUSION_FAILED_LATENCY_PERCENTILE: return "failed to to a high latency in the 99.99%'ile";
 		case CONCLUSION_PASSED: return "passed";
 		case CONCLUSION_SCAN_PROBLEM: return "scan_problem";
 		case CONCLUSION_ABORTED: return "scan_aborted";
@@ -71,6 +103,172 @@ enum scan_mode str_to_scan_mode(const char *s)
 	return SCAN_MODE_UNKNOWN;
 }
 
+static void disk_ata_monitor_start(disk_t *disk)
+{
+	if (disk_smart_trip(&disk->dev) == 1) {
+		ERROR("Disk has a SMART TRIP at the start of the test, it should be discarded anyhow");
+		disk->state.ata.is_smart_tripped = true;
+	} else {
+		disk->state.ata.is_smart_tripped = false;
+	}
+
+	disk->state.ata.smart_table = smart_table_for_disk(disk->vendor, disk->model, disk->fw_rev);
+	if (disk->state.ata.smart_table == NULL)
+		ERROR("BUG! Failed to setup smart table for the disk.");
+
+	disk->state.ata.smart_num = disk_smart_attributes(&disk->dev, disk->state.ata.smart, ARRAY_SIZE(disk->state.ata.smart));
+
+	if (disk->state.ata.smart_num > 0) {
+		// First look at temperatures
+		int min_temp = -1;
+		int max_temp = -1;
+		int temp = ata_smart_get_temperature(disk->state.ata.smart, disk->state.ata.smart_num, disk->state.ata.smart_table, &min_temp, &max_temp);
+		disk->state.ata.last_temp = temp;
+
+		if (min_temp > 0 || max_temp > 0)
+			INFO("Disk start temperature is %d (lifetime min %d and lifetime max %d)", temp, min_temp, max_temp);
+		else
+			INFO("Disk start temperature is %d", temp);
+
+		// First look on reallocations
+		disk->state.ata.last_reallocs = ata_smart_get_num_reallocations(disk->state.ata.smart, disk->state.ata.smart_num,
+				disk->state.ata.smart_table);
+		disk->state.ata.last_pending_reallocs = ata_smart_get_num_pending_reallocations(disk->state.ata.smart, disk->state.ata.smart_num,
+				disk->state.ata.smart_table);
+
+		// Now take a first look at the CRC error counters
+		disk->state.ata.last_crc_errors = ata_smart_get_num_crc_errors(disk->state.ata.smart, disk->state.ata.smart_num,
+				disk->state.ata.smart_table);
+	} else {
+		ERROR("Failed to read SMART attributes from device");
+	}
+}
+
+static void ata_test_temp(disk_t *disk, ata_smart_attr_t *smart, int smart_num)
+{
+	int min_temp = -1;
+	int max_temp = -1;
+	int temp = ata_smart_get_temperature(smart, smart_num, disk->state.ata.smart_table, &min_temp, &max_temp);
+
+	if (temp != disk->state.ata.last_temp) {
+		INFO("Disk temperature changed from %d to %d", disk->state.ata.last_temp, temp);
+		disk->state.ata.last_temp = temp;
+	}
+
+	if (temp >= TEMP_THRESHOLD) {
+		spinner_t spinner;
+		INFO("Pausing scan due to high disk temperature");
+		spinner_init(&spinner);
+		while (temp >= TEMP_THRESHOLD) {
+			sleep(1);
+			spinner_update(&spinner);
+			smart_num = disk_smart_attributes(&disk->dev, smart, ARRAY_SIZE(smart));
+			if (smart_num > 0) {
+				temp = ata_smart_get_temperature(smart, smart_num, disk->state.ata.smart_table, &min_temp, &max_temp);
+			} else {
+				ERROR("Failed to read temperature while paused!");
+				break;
+			}
+		}
+		spinner_done();
+		INFO("Finished pause, temperature is now %d", temp);
+	}
+}
+
+static void ata_test_reallocs(disk_t *disk, ata_smart_attr_t *smart, int smart_num)
+{
+	int num_reallocs;
+	int num_pending_reallocs;
+
+	num_reallocs = ata_smart_get_num_reallocations(smart, smart_num, disk->state.ata.smart_table);
+	num_pending_reallocs = ata_smart_get_num_pending_reallocations(smart, smart_num, disk->state.ata.smart_table);
+
+	if (num_reallocs > disk->state.ata.last_reallocs) {
+		INFO("Number of reallocated sectors increased from %d to %d\n", disk->state.ata.last_reallocs, num_reallocs);
+		disk->state.ata.last_reallocs = num_reallocs;
+	}
+
+	if (num_pending_reallocs != disk->state.ata.last_pending_reallocs) {
+		INFO("Number of pending sectors for reallocations changed from %d to %d\n", disk->state.ata.last_pending_reallocs,
+				num_pending_reallocs);
+		disk->state.ata.last_pending_reallocs = num_pending_reallocs;
+	}
+}
+
+static void ata_test_crc_errors(disk_t *disk, ata_smart_attr_t *smart, int smart_num)
+{
+	int crc_errors;
+
+	crc_errors = ata_smart_get_num_crc_errors(smart, smart_num, disk->state.ata.smart_table);
+	if (crc_errors != disk->state.ata.last_crc_errors) {
+		ERROR("CRC errors increased from %d to %d, your problem is not the disk but in a cable most likely!",
+				disk->state.ata.last_crc_errors, crc_errors);
+		disk->state.ata.last_crc_errors = crc_errors;
+	}
+}
+
+static void disk_ata_monitor(disk_t *disk)
+{
+	ata_smart_attr_t smart[MAX_SMART_ATTRS];
+	int smart_num;
+
+	if (!disk->state.ata.is_smart_tripped && disk_smart_trip(&disk->dev) == 1) {
+		ERROR("Disk has a SMART TRIP in the middle of the test, it should be discarded!");
+		disk->state.ata.is_smart_tripped = true;
+	}
+
+	smart_num = disk_smart_attributes(&disk->dev, smart, ARRAY_SIZE(smart));
+
+	if (smart_num > 0) {
+		ata_test_temp(disk, smart, smart_num);
+		ata_test_reallocs(disk, smart, smart_num);
+		ata_test_crc_errors(disk, smart, smart_num);
+	} else {
+		ERROR("Failed to read SMART attributes from device");
+	}
+}
+
+static void disk_ata_monitor_end(disk_t *disk)
+{
+	ata_smart_attr_t smart[MAX_SMART_ATTRS];
+	int smart_num;
+	int num_reallocs;
+	int num_pending_reallocs;
+
+	if (disk_smart_trip(&disk->dev) == 1) {
+		ERROR("Disk has a SMART TRIP at the end of the test, it should be discarded!");
+	} else if (disk->state.ata.is_smart_tripped) {
+		ERROR("Disk had a SMART TRIP during the test but it disappeared. This is super weird!!!");
+	}
+
+	smart_num = disk_smart_attributes(&disk->dev, smart, ARRAY_SIZE(smart));
+	num_reallocs = ata_smart_get_num_reallocations(smart, smart_num, disk->state.ata.smart_table);
+	num_pending_reallocs = ata_smart_get_num_pending_reallocations(smart, smart_num, disk->state.ata.smart_table);
+
+	if (num_pending_reallocs > 0) {
+		INFO("At the end of the test there are still some sectors pending reallocation, this is rather unexpected but can be lived with.");
+	}
+
+	if (num_reallocs > 1000) {
+		INFO("Number of reallocated sectors is above 1000, you should probably stop using this disk!");
+	}
+}
+
+static void disk_scsi_monitor_start(disk_t *disk)
+{
+	(void)disk;
+}
+
+static void disk_scsi_monitor(disk_t *disk)
+{
+	(void)disk;
+}
+
+static void disk_scsi_monitor_end(disk_t *disk)
+{
+	(void)disk;
+}
+
 int disk_open(disk_t *disk, const char *path, int fix, unsigned latency_graph_len)
 {
 	memset(disk, 0, sizeof(*disk));
@@ -123,6 +321,8 @@ int disk_open(disk_t *disk, const char *path, int fix, unsigned latency_graph_le
 	strncpy(disk->path, path, sizeof(disk->path));
 	disk->path[sizeof(disk->path)-1] = 0;
 
+	hdr_init(1, 60*1000*1000, 3, &disk->histogram);
+
 	disk->latency_graph_len = latency_graph_len;
 	disk->latency_graph = calloc(latency_graph_len, sizeof(latency_t));
 	if (disk->latency_graph == NULL) {
@@ -130,6 +330,11 @@ int disk_open(disk_t *disk, const char *path, int fix, unsigned latency_graph_le
 		goto Error;
 	}
 
+	if (disk->is_ata)
+		disk_ata_monitor_start(disk);
+	else
+		disk_scsi_monitor_start(disk);
+
 	INFO("Opened disk %s sector size %"PRIu64" num bytes %"PRIu64, path, disk->sector_size, disk->num_bytes);
 	return 0;
 
@@ -140,6 +345,11 @@ Error:
 
 int disk_close(disk_t *disk)
 {
+	if (disk->is_ata)
+		disk_ata_monitor_end(disk);
+	else
+		disk_scsi_monitor_end(disk);
+
 	INFO("Closed disk %s", disk->path);
 	disk_dev_close(&disk->dev);
 	if (disk->latency_graph) {
@@ -287,12 +497,7 @@ static bool disk_scan_part(disk_t *disk, uint64_t offset, void *data, int data_s
 		report_scan_success(disk, offset, data_size, t);
 	}
 
-	unsigned hist_idx = 0;
-	while (t_msec >= histogram_time[hist_idx].top_val && hist_idx < ARRAY_SIZE(disk->histogram) - 1) {
-		hist_idx++;
-	}
-	disk->histogram[hist_idx]++;
-
+	hdr_record_value(disk->histogram, t / 1000);
 	latency_bucket_add(disk, t_msec, state);
 
 	if (t_msec > 1000) {
@@ -428,26 +633,14 @@ static void set_realtime(bool realtime)
 
 static enum conclusion conclusion_calc(disk_t *disk)
 {
-	unsigned i;
-	uint64_t total;
-
 	if (disk->num_errors > 0)
-		return CONCLUSION_FAILED;
+		return CONCLUSION_FAILED_IO_ERRORS;
 
-	for (i = 0, total = 0; i < ARRAY_SIZE(histogram_time); i++) {
-		total += disk->histogram[i];
-	}
+	if (hdr_max(disk->histogram) > 10000000)
+		return CONCLUSION_FAILED_MAX_LATENCY;
 
-	const double total_d = (double)total;
-	for (i = 0; i < ARRAY_SIZE(histogram_time); i++) {
-		double percent = ((double)disk->histogram[i]) / total_d;
-		if (percent > histogram_time[i].percent_bad) {
-			const uint64_t low_end = i == 0 ? 0 : histogram_time[i-1].top_val;
-			const uint64_t high_end = histogram_time[i].top_val;
-			ERROR("Disk has %"PRIu64" (%f%%) IOs from %"PRIu64" to %"PRIu64" msecs, it is a bad disk!", disk->histogram[i], percent, low_end, high_end);
-			return CONCLUSION_FAILED;
-		}
-	}
+	if (hdr_value_at_percentile(disk->histogram, 99.99) > 8000000)
+		return CONCLUSION_FAILED_LATENCY_PERCENTILE;
 
 	VERBOSE("Disk has passed the test");
 	return CONCLUSION_PASSED;
@@ -505,6 +698,7 @@ int disk_scan(disk_t *disk, enum scan_mode mode, unsigned data_size)
 		goto Exit;
 	}
 
+	verbose_extra_newline = 1;
 	for (offset = 0; disk->run && offset < disk_size_bytes; offset += latency_stride * disk->sector_size) {
 		VERBOSE("Scanning stride starting at %"PRIu64" done %"PRIu64"%%", offset, offset*100/disk_size_bytes);
 		progress_calc(disk, &state, 0);
@@ -512,7 +706,13 @@ int disk_scan(disk_t *disk, enum scan_mode mode, unsigned data_size)
 		if (!disk_scan_latency_stride(disk, &state, offset, data_size, scan_order))
 			break;
 		latency_bucket_finish(disk, &state, offset + latency_stride * disk->sector_size);
+
+		if (disk->is_ata)
+			disk_ata_monitor(disk);
+		else
+			disk_scsi_monitor(disk);
 	}
+	verbose_extra_newline = 0;
 
 	if (!disk->run) {
 		INFO("Disk scan interrupted");
